#import "JubGtkRosterUI.h"
#import "JubGObjectMap.h"
#import "JubGtkChatUI.h"

static gboolean filter_roster_by_presence(GtkTreeModel *model,
    GtkTreeIter *iter, gpointer data)
{
	char *jid_s;
	OFString *jid;
	OFCountedSet *presences = data;

	gtk_tree_model_get(model, iter, 1, &jid_s, -1);

	// Groups have no JID
	if (!jid_s)
		return TRUE;

	jid = [[OFString alloc] initWithUTF8String: jid_s];

	g_free(jid_s);

	int num = [presences countForObject: jid];
	if (num) {
		[jid release];
		return TRUE;
	} else {
		[jid release];
		return FALSE;
	}
}

@implementation JubGtkRosterUI
- initWithBuilder: (GtkBuilder*)builder_
{
	self = [super init];

	@try {
		groupMap = [[OFMapTable alloc]
		    initWithKeyFunctions: keyFunctions
			  valueFunctions: rowRefFunctions];
		contactMap = [[OFMutableDictionary alloc] init];
		chatMap = [[OFMutableDictionary alloc] init];
		presences = [[OFCountedSet alloc] init];

		builder = g_object_ref(builder_);

		roster_model = GTK_TREE_STORE(gtk_builder_get_object(builder,
		    "RosterTreeStore"));

		roster_filter = GTK_TREE_MODEL_FILTER(
		    gtk_builder_get_object(builder, "RosterTreeModelFilter"));

		gtk_tree_model_filter_set_visible_func(roster_filter,
		    filter_roster_by_presence, presences, NULL);
	} @catch (id e) {
		[self release];
		@throw e;
	}

	return self;
}

- (void)dealloc
{
	[groupMap release];
	[contactMap release];
	[presences release];

	if (roster_model)
		g_object_unref(roster_model);

	if (roster_filter)
		g_object_unref(roster_filter);

	g_object_unref(builder);

	[super dealloc];
}

/* Presence handling */
static gboolean refilter_roster(gpointer data)
{
	gtk_tree_model_filter_refilter(data);

	return FALSE;
}

-   (void)connection: (XMPPConnection*)connection
  didReceivePresence: (XMPPPresence*)presence
{
	if ([presence.type isEqual: @"available"])
		[presences addObject: [presence.from bareJID]];
	else if ([presence.type isEqual: @"unavailable"])
		[presences removeObject: [presence.from bareJID]];

	g_idle_add(refilter_roster, roster_filter);
}

// FIXME: This needs to move somewhere else
-  (void)connection: (XMPPConnection*)connection
  didReceiveMessage: (XMPPMessage*)message
{
	JubGtkChatUI *chat =
	    [chatMap objectForKey: [message.from bareJID]];
	if (chat == nil) {
		OFString * title = [@"Chat with " stringByAppendingString:
		    [message.from bareJID]];

		chat = [[[JubGtkChatUI alloc]
		    initWithTitle: title
		       closeBlock: ^{
				[chatMap removeObjectForKey:
				    [message.from bareJID]];
			}
			sendBlock: ^(OFString *text) {
				XMPPMessage *msg =
				    [XMPPMessage messageWithType: @"chat"];
				msg.to = message.from;
				msg.body = text;
				[connection sendStanza: msg];
			}
		] autorelease];

		[chatMap setObject: chat
			    forKey: [message.from bareJID]];
	}

	[chat addMessage: message.body
		  sender: [message.from bareJID]];
}

/* Roster Delegate methods */
struct add_roster_item_param {
	OFString *group;
	OFString *name;
	XMPPJID *jid;
	OFMapTable *groupMap;
	OFMapTable *contactRows;
	GtkTreeStore *roster_model;
};

static gboolean add_roster_item(gpointer user_data)
{
	struct add_roster_item_param *params = user_data;
	GtkTreeIter group_iter, contact_iter;
	GtkTreeRowReference *group_ref, *contact_ref;
	GtkTreePath *group_path, *contact_path;

	group_ref = [params->groupMap valueForKey: params->group];

	if (!group_ref) {
		// Create new group row
		gtk_tree_store_append(params->roster_model, &group_iter, NULL);
		gtk_tree_store_set(params->roster_model, &group_iter,
		    0, [params->group UTF8String], -1);

		group_path = gtk_tree_model_get_path(GTK_TREE_MODEL(
			params->roster_model), &group_iter);

		group_ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(
			params->roster_model), group_path);

		[params->groupMap setValue: group_ref
				    forKey: params->group];
	} else {
		// Get iter for existing group row
		group_path = gtk_tree_row_reference_get_path(group_ref);

		gtk_tree_model_get_iter(GTK_TREE_MODEL(params->roster_model),
		    &group_iter, group_path);
	}
	gtk_tree_path_free(group_path);

	// Create new contact row
	gtk_tree_store_append(params->roster_model, &contact_iter, &group_iter);
	if (params->name)
		gtk_tree_store_set(params->roster_model, &contact_iter,
		    0, [params->name UTF8String],
		    1, [[params->jid bareJID] UTF8String], -1);
	else
		gtk_tree_store_set(params->roster_model, &contact_iter,
		    0, [params->jid.node UTF8String],
		    1, [[params->jid bareJID] UTF8String], -1);

	contact_path = gtk_tree_model_get_path(GTK_TREE_MODEL(
		params->roster_model), &contact_iter);

	contact_ref = gtk_tree_row_reference_new(GTK_TREE_MODEL(
		params->roster_model), contact_path);

	gtk_tree_path_free(contact_path);

	[params->contactRows setValue: contact_ref
			       forKey: params->group];

	[params->group release];
	[params->name release];
	[params->jid release];
	[params->groupMap release];
	[params->contactRows release];
	g_object_unref(params->roster_model);
	free(params);

	return FALSE;
}

struct remove_roster_item_param {
	OFString *group;
	OFMapTable *groupMap;
	OFMapTable *contactRows;
	GtkTreeStore *roster_model;
};

static gboolean remove_roster_item(gpointer user_data)
{
	struct remove_roster_item_param *params = user_data;
	GtkTreeIter contact_iter, group_iter;
	GtkTreePath *contact_path, *group_path;
	GtkTreeRowReference *contact_ref, *group_ref;

	contact_ref = [params->contactRows valueForKey: params->group];
	contact_path = gtk_tree_row_reference_get_path(contact_ref);
	gtk_tree_model_get_iter(GTK_TREE_MODEL(params->roster_model),
	    &contact_iter, contact_path);

	gtk_tree_store_remove(params->roster_model, &contact_iter);

	group_ref = [params->groupMap valueForKey: params->group];
	group_path = gtk_tree_row_reference_get_path(group_ref);
	gtk_tree_model_get_iter(GTK_TREE_MODEL(params->roster_model),
	    &group_iter, group_path);

	if (!gtk_tree_model_iter_has_child(GTK_TREE_MODEL(params->roster_model),
	    &group_iter)) {
		gtk_tree_store_remove(params->roster_model, &group_iter);
		[params->groupMap removeValueForKey: params->group];
	}

	gtk_tree_path_free(group_path);

	[params->group release];
	[params->groupMap release];
	[params->contactRows release];
	g_object_unref(params->roster_model);
	free(params);

	return FALSE;
}

- (void)rosterWasReceived: (XMPPRoster*)roster_
{
	of_log(@"Handling roster");
	[[roster_ rosterItems] enumerateKeysAndObjectsUsingBlock:
	    ^(OFString *bareJID, XMPPRosterItem *item, BOOL *stop) {
		OFArray *groups;
		OFMapTable *contactRows = [OFMapTable
		    mapTableWithKeyFunctions: keyFunctions
			      valueFunctions: rowRefFunctions];

		[contactMap setObject: contactRows
			       forKey: bareJID];

		if (item.groups != nil)
			groups = item.groups;
		else
			groups = @[@"General"];

		for (OFString *group in groups) {
			struct add_roster_item_param *params =
			    malloc(sizeof(*params));
			params->group = [group retain];
			params->name = [item.name retain];
			params->jid = [item.JID retain];
			params->groupMap = [groupMap retain];
			params->contactRows = [contactRows retain];
			params->roster_model = g_object_ref(roster_model);
			g_idle_add(add_roster_item, params);
		}
	}];
}

-         (void)roster: (XMPPRoster*)roster_
  didReceiveRosterItem: (XMPPRosterItem*)item
{
	OFArray *groups;
	XMPPRosterItem *oldItem =
	    [roster_.rosterItems objectForKey: [item.JID bareJID]];

	if (oldItem) {
		if (oldItem.groups != nil)
			groups = oldItem.groups;
		else
			groups = @[@"General"];

		for (OFString *group in groups) {
			struct remove_roster_item_param *params =
			    malloc(sizeof(*params));
			params->group = [group retain];
			params->contactRows = [[contactMap objectForKey:
			    [oldItem.JID bareJID]] retain];
			params->groupMap = [groupMap retain];
			params->roster_model = g_object_ref(roster_model);
			g_idle_add(remove_roster_item, params);
		}

		[contactMap removeObjectForKey: [item.JID bareJID]];
	}

	if (![item.subscription isEqual: @"remove"]) {
		OFMapTable *contactRows = [OFMapTable
		    mapTableWithKeyFunctions: keyFunctions
			      valueFunctions: rowRefFunctions];

		[contactMap setObject: contactRows
			       forKey: [item.JID bareJID]];

		if (item.groups != nil)
			groups = item.groups;
		else
			groups = @[@"General"];

		for (OFString *group in groups) {
			struct add_roster_item_param *params =
			    malloc(sizeof(*params));
			params->group = [group retain];
			params->name = [item.name retain];
			params->jid = [item.JID retain];
			params->groupMap = [groupMap retain];
			params->contactRows = [contactRows retain];
			params->roster_model = g_object_ref(roster_model);
			g_idle_add(add_roster_item, params);
		}
	}
}

@end
